/**
 * Copyright (c) 2017-2020, NVIDIA CORPORATION.  All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <string.h>
#include <string>
#include <sstream>
#include <iostream>
#include <ostream>
#include <fstream>
#include "gstdsexample.h"
#include <sys/time.h>
#include <utility> 
#include <nvbufsurftransform.h>
#include <nvbufsurface.h>
#include <nvll_osd_api.h>
GST_DEBUG_CATEGORY_STATIC (gst_dsexample_debug);
#define GST_CAT_DEFAULT gst_dsexample_debug
#define USE_EGLIMAGE 1
#define PGIE_CLASS_ID_VEHICLE 0
#define PGIE_CLASS_ID_MOTORBIKE 1
#define PGIE_CLASS_ID_PERSON 2
/* enable to write transformed cvmat to files */
/* #define DSEXAMPLE_DEBUG */
static GQuark _dsmeta_quark = 0;
static guint counter = 0;
/* Enum to identify properties */
enum
{
  PROP_0,
  PROP_UNIQUE_ID,
  PROP_PROCESSING_WIDTH,
  PROP_PROCESSING_HEIGHT,
  PROP_PROCESS_FULL_FRAME,
  PROP_BLUR_OBJECTS,
  PROP_GPU_DEVICE_ID
};

#define CHECK_NVDS_MEMORY_AND_GPUID(object, surface)  \
  ({ int _errtype=0;\
   do {  \
    if ((surface->memType == NVBUF_MEM_DEFAULT || surface->memType == NVBUF_MEM_CUDA_DEVICE) && \
        (surface->gpuId != object->gpu_id))  { \
    GST_ELEMENT_ERROR (object, RESOURCE, FAILED, \
        ("Input surface gpu-id doesnt match with configured gpu-id for element," \
         " please allocate input using unified memory, or use same gpu-ids"),\
        ("surface-gpu-id=%d,%s-gpu-id=%d",surface->gpuId,GST_ELEMENT_NAME(object),\
         object->gpu_id)); \
    _errtype = 1;\
    } \
    } while(0); \
    _errtype; \
  })


/* Default values for properties */
#define DEFAULT_UNIQUE_ID 15
#define DEFAULT_PROCESSING_WIDTH 640
#define DEFAULT_PROCESSING_HEIGHT 480
#define DEFAULT_PROCESS_FULL_FRAME TRUE
#define DEFAULT_BLUR_OBJECTS FALSE
#define DEFAULT_GPU_ID 0

#define RGB_BYTES_PER_PIXEL 3
#define RGBA_BYTES_PER_PIXEL 4
#define Y_BYTES_PER_PIXEL 1
#define UV_BYTES_PER_PIXEL 2

#define MIN_INPUT_OBJECT_WIDTH 16
#define MIN_INPUT_OBJECT_HEIGHT 16

#define CHECK_NPP_STATUS(npp_status,error_str) do { \
  if ((npp_status) != NPP_SUCCESS) { \
    g_print ("Error: %s in %s at line %d: NPP Error %d\n", \
        error_str, __FILE__, __LINE__, npp_status); \
    goto error; \
  } \
} while (0)

#define CHECK_CUDA_STATUS(cuda_status,error_str) do { \
  if ((cuda_status) != cudaSuccess) { \
    g_print ("Error: %s in %s at line %d (%s)\n", \
        error_str, __FILE__, __LINE__, cudaGetErrorName(cuda_status)); \
    goto error; \
  } \
} while (0)

/* By default NVIDIA Hardware allocated memory flows through the pipeline. We
 * will be processing on this type of memory only. */
#define GST_CAPS_FEATURE_MEMORY_NVMM "memory:NVMM"
static GstStaticPadTemplate gst_dsexample_sink_template =
GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE_WITH_FEATURES
        (GST_CAPS_FEATURE_MEMORY_NVMM,
            "{ NV12, RGBA, I420 }")));

static GstStaticPadTemplate gst_dsexample_src_template =
GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE_WITH_FEATURES
        (GST_CAPS_FEATURE_MEMORY_NVMM,
            "{ NV12, RGBA, I420 }")));

/* Define our element type. Standard GObject/GStreamer boilerplate stuff */
#define gst_dsexample_parent_class parent_class
G_DEFINE_TYPE (GstDsExample, gst_dsexample, GST_TYPE_BASE_TRANSFORM);

static void gst_dsexample_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_dsexample_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_dsexample_set_caps (GstBaseTransform * btrans,
    GstCaps * incaps, GstCaps * outcaps);
static gboolean gst_dsexample_start (GstBaseTransform * btrans);
static gboolean gst_dsexample_stop (GstBaseTransform * btrans);

static GstFlowReturn gst_dsexample_transform_ip (GstBaseTransform *
    btrans, GstBuffer * inbuf);

static void
attach_metadata_full_frame (GstDsExample * dsexample, NvDsFrameMeta *frame_meta,
    gdouble scale_ratio, DsExampleOutput * output, guint batch_id);
static void attach_metadata_object (GstDsExample * dsexample,
    NvDsObjectMeta * obj_meta, DsExampleOutput * output);

/* Install properties, set sink and src pad capabilities, override the required
 * functions of the base class, These are common to all instances of the
 * element.
 */
static void
gst_dsexample_class_init (GstDsExampleClass * klass){
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseTransformClass *gstbasetransform_class;

  /* Indicates we want to use DS buf api */
  g_setenv ("DS_NEW_BUFAPI", "1", TRUE);

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstbasetransform_class = (GstBaseTransformClass *) klass;

  /* Overide base class functions */
  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_dsexample_set_property);
  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_dsexample_get_property);

  gstbasetransform_class->set_caps = GST_DEBUG_FUNCPTR (gst_dsexample_set_caps);
  gstbasetransform_class->start = GST_DEBUG_FUNCPTR (gst_dsexample_start);
  gstbasetransform_class->stop = GST_DEBUG_FUNCPTR (gst_dsexample_stop);

  gstbasetransform_class->transform_ip =
      GST_DEBUG_FUNCPTR (gst_dsexample_transform_ip);

  /* Install properties */
  g_object_class_install_property (gobject_class, PROP_UNIQUE_ID,
      g_param_spec_uint ("unique-id",
          "Unique ID",
          "Unique ID for the element. Can be used to identify output of the"
          " element", 0, G_MAXUINT, DEFAULT_UNIQUE_ID, (GParamFlags)
          (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));

  g_object_class_install_property (gobject_class, PROP_PROCESSING_WIDTH,
      g_param_spec_int ("processing-width",
          "Processing Width",
          "Width of the input buffer to algorithm",
          1, G_MAXINT, DEFAULT_PROCESSING_WIDTH, (GParamFlags)
          (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));

  g_object_class_install_property (gobject_class, PROP_PROCESSING_HEIGHT,
      g_param_spec_int ("processing-height",
          "Processing Height",
          "Height of the input buffer to algorithm",
          1, G_MAXINT, DEFAULT_PROCESSING_HEIGHT, (GParamFlags)
          (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));

  g_object_class_install_property (gobject_class, PROP_PROCESS_FULL_FRAME,
      g_param_spec_boolean ("full-frame",
          "Full frame",
          "Enable to process full frame or disable to process objects detected"
          "by primary detector", DEFAULT_PROCESS_FULL_FRAME, (GParamFlags)
          (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));

  g_object_class_install_property (gobject_class, PROP_BLUR_OBJECTS,
      g_param_spec_boolean ("blur-objects",
          "Blur Objects",
          "Enable to blur the objects detected in full-frame=0 mode"
          "by primary detector", DEFAULT_BLUR_OBJECTS, (GParamFlags)
          (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));

  g_object_class_install_property (gobject_class, PROP_GPU_DEVICE_ID,
      g_param_spec_uint ("gpu-id",
          "Set GPU Device ID",
          "Set GPU Device ID", 0,
          G_MAXUINT, 0,
          GParamFlags
          (G_PARAM_READWRITE |
              G_PARAM_STATIC_STRINGS | GST_PARAM_MUTABLE_READY)));
  /* Set sink and src pad capabilities */
  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&gst_dsexample_src_template));
  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&gst_dsexample_sink_template));

  /* Set metadata describing the element */
  gst_element_class_set_details_simple (gstelement_class,
      "DsExample plugin",
      "DsExample Plugin",
      "Process a 3rdparty example algorithm on objects / full frame",
      "NVIDIA Corporation. Post on Deepstream for Tesla forum for any queries "
      "@ https://devtalk.nvidia.com/default/board/209/");
}

static void
gst_dsexample_init (GstDsExample * dsexample)
{
  GstBaseTransform *btrans = GST_BASE_TRANSFORM (dsexample);

  /* We will not be generating a new buffer. Just adding / updating
   * metadata. */
  gst_base_transform_set_in_place (GST_BASE_TRANSFORM (btrans), TRUE);
  /* We do not want to change the input caps. Set to passthrough. transform_ip
   * is still called. */
  gst_base_transform_set_passthrough (GST_BASE_TRANSFORM (btrans), TRUE);

  /* Initialize all property variables to default values */
  dsexample->unique_id = DEFAULT_UNIQUE_ID;
  dsexample->processing_width = DEFAULT_PROCESSING_WIDTH;
  dsexample->processing_height = DEFAULT_PROCESSING_HEIGHT;
  dsexample->process_full_frame = DEFAULT_PROCESS_FULL_FRAME;
  dsexample->blur_objects = DEFAULT_BLUR_OBJECTS;
  dsexample->gpu_id = DEFAULT_GPU_ID;
  
  /* This quark is required to identify NvDsMeta when iterating through
   * the buffer metadatas */
  if (!_dsmeta_quark)
    _dsmeta_quark = g_quark_from_static_string (NVDS_META_STRING);
}

/* Function called when a property of the element is set. Standard boilerplate.
 */
static void
gst_dsexample_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstDsExample *dsexample = GST_DSEXAMPLE (object);
  switch (prop_id) {
    case PROP_UNIQUE_ID:
      dsexample->unique_id = g_value_get_uint (value);
      break;
    case PROP_PROCESSING_WIDTH:
      dsexample->processing_width = g_value_get_int (value);
      break;
    case PROP_PROCESSING_HEIGHT:
      dsexample->processing_height = g_value_get_int (value);
      break;
    case PROP_PROCESS_FULL_FRAME:
      dsexample->process_full_frame = g_value_get_boolean (value);
      break;
    case PROP_BLUR_OBJECTS:
      dsexample->blur_objects = g_value_get_boolean (value);
      break;
    case PROP_GPU_DEVICE_ID:
      dsexample->gpu_id = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* Function called when a property of the element is requested. Standard
 * boilerplate.
 */
static void
gst_dsexample_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstDsExample *dsexample = GST_DSEXAMPLE (object);

  switch (prop_id) {
    case PROP_UNIQUE_ID:
      g_value_set_uint (value, dsexample->unique_id);
      break;
    case PROP_PROCESSING_WIDTH:
      g_value_set_int (value, dsexample->processing_width);
      break;
    case PROP_PROCESSING_HEIGHT:
      g_value_set_int (value, dsexample->processing_height);
      break;
    case PROP_PROCESS_FULL_FRAME:
      g_value_set_boolean (value, dsexample->process_full_frame);
      break;
    case PROP_BLUR_OBJECTS:
      g_value_set_boolean (value, dsexample->blur_objects);
      break;
    case PROP_GPU_DEVICE_ID:
      g_value_set_uint (value, dsexample->gpu_id);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}
#ifdef ENABLE_MAPLETREE
/**
Thread
**/
returnfromillegalbox gettimeellapseforilligalparking(guint64 id){
   returnfromillegalbox ret; ret.elapsedtime=0; ret.found=false;
   struct timeval currenttime;
   gettimeofday(&currenttime, NULL);
   for(std::vector<box>::iterator it=illegalparked_boxes.begin(); it != illegalparked_boxes.end(); it++){
       if(it->id == id){
          //printf("found %lu searched %lu\n",it->id,id);
          ret.elapsedtime = currenttime.tv_sec - it->start_tv.tv_sec;
          gettimeofday(&it->lastupdated_tv, NULL);
          ret.found=true;
          return ret;
       }
   }
   int count=0;
   gettimeofday(&currenttime, NULL);
   uint64_t longestelapsedtime=0;
   while(count < illegalparked_boxes.size()){
       uint64_t elapsedtime = currenttime.tv_sec - illegalparked_boxes[count].lastupdated_tv.tv_sec;
       uint64_t temp=currenttime.tv_sec - illegalparked_boxes[count].start_tv.tv_sec;
       if(temp > longestelapsedtime)
         longestelapsedtime = temp;
       if(elapsedtime>2){
          illegalparked_boxes.erase(illegalparked_boxes.begin()+count);
          count=count-1;
       }
       count=count+1;
   } 
   if(longestelapsedtime < TIMEINTERVAL)
   {
      std::lock_guard<std::mutex> guard(alert_lock);
      alertsoruce [0] = 0; alertsoruce [1] = 0; alertsoruce [2] = 0; alertsoruce [3] = 0;
   }
   return ret;
}
void* func(void* arg)
{
    pthread_detach(pthread_self());
    while(threadrunning){
       box getbox;
       {
          std::lock_guard<std::mutex> guard(m_lock);
          if(!boxes.empty()){
             //printf("vector size %d \n", boxes.size());
             getbox = boxes.back();
             boxes.pop_back();
          }
       }
       if(getbox.source_num>=0 && getbox.source_num<=8 ){                         
             for (int source=0; source < numberofsources; source++){
                
                if(getbox.source_num == refboxes_calculate[source].source_num){	 
                        //g_print("box source %d refboxe source %d and size %d \n", getbox.source_num, refboxes_calculate[source].source_num, refboxes_calculate[source].size);      			
		        for(int boxidx=0; boxidx < refboxes_calculate[source].size; boxidx=boxidx+4){
                            int x_center = (getbox.x_max - getbox.x_min) + getbox.x_min;
                            int y_center = (getbox.y_max - getbox.y_min) + getbox.y_min;
		            if(x_center > refboxes_calculate[source].corners[boxidx+0] && x_center < refboxes_calculate[source].corners[boxidx+2] && y_center >    
                               refboxes_calculate[source].corners[boxidx+1] && y_center < refboxes_calculate[source].corners[boxidx+3]){
                                 //g_print("1source %d %d %d %d \n",x_ref_min, y_ref_min, x_ref_max, y_ref_max);
                                 //g_print("1box %d %lu %d %d %d %d %ld\n",getbox.source_num, getbox.id, getbox.x_min, getbox.y_min, getbox.x_max, getbox.y_max, getbox.start_tv.tv_sec); 
                                 returnfromillegalbox ret_ = gettimeellapseforilligalparking(getbox.id);
                                 //printf("rettimeinterval %lu \n", ret_.elapsedtime);
                                 if(ret_.found==true && ret_.elapsedtime>TIMEINTERVAL){
                                    std::lock_guard<std::mutex> guard(alert_lock);
                                    alertsoruce [getbox.source_num] = 1;
                                 }else if( ret_.found==false ){
                                    illegalparked_boxes.push_back(getbox);
                                    //printf("inserted %lu illegalparked_boxes size %d for source %d\n",getbox.id , illegalparked_boxes.size(), refboxes_calculate[source].source_num);
                                 }   
                                 //g_print("exit from  boxidx \n"); 
                                 break;
                            }
		        }
                        //g_print("exit from  source \n");
                        break;//exit from for (int source=0; source < numberofsources; source++)
                }
	     } 
       }
    }
    printf("thread loop stops\n");
    // exit the current thread
    pthread_exit(NULL);
}
int display_fileread(const char* filename){
  char * line = NULL;
  size_t len = 0;
  ssize_t read;
  FILE *fptr;  
  numberofsources=0;
  fptr = fopen(filename, "r");
  if(fptr == NULL)
  {
     g_print("File read error!\n"); 
     return -1;
  }
  
  //set reference boxes
  
  int linenum=-1;
  while ((read = getline(&line, &len, fptr)) != -1) {
     char * token = strtok(line, ":");
     int count=0;
     char * current;
     while( token != NULL ) { 
        if(count==0 ){
          current=token;          
        }
        if(strcmp(current, "numsources") == 0 && count>0)
        {
            numberofsources = atoi(token);            
            refboxes_display = (refbox*)malloc(numberofsources * sizeof(refbox)); 
           
        }   
        else if(strstr(current, "source") != NULL && count>0)
        {
           //printf("\nlinenum %d\n",linenum);
           //printf("current %s\n",current);
           char * source1 = strtok(current, "source"); 
           //printf("source1 %s\n",source1);
           char * source2 = strtok(source1, "-");           
           refboxes_display[linenum].source_num = atoi(source2)-1;
           //printf("source_num %d\n", refboxes_display[linenum].source_num);
           while( source2 != NULL ) { 
              source2 = strtok(NULL, " ");
              refboxes_display[linenum].size = atoi(source2)*4;
              refboxes_display[linenum].corners = (int*)malloc(atoi(source2)*4 * sizeof(int));
              source2 = strtok(NULL, " ");
           }
           int countcorners=0;
           //printf("\ntoken %s\n",token);
	   do{  
		int l = strcspn (token, ",");
                //printf("countcorners %d num %d ",countcorners, atoi(token));
		refboxes_display[linenum].corners[countcorners] = atoi(token);  
		token += l + 1; 
		countcorners = countcorners+1; 		      
	  }while(token[-1]);
        }        
          
        token = strtok(NULL, " ");
        count=count+1;
     }
     linenum=linenum+1;
  }
  
  fclose(fptr);
  printf("numberofsources %d\n ", numberofsources);
  for(linenum=0; linenum < numberofsources; linenum++){
     printf( "\nsources %d size %d\n", refboxes_display[linenum].source_num, refboxes_display[linenum].size );
     for(int countcorners=0; countcorners < refboxes_display[linenum].size; countcorners++)
       printf( "%d ", refboxes_display[linenum].corners[countcorners] );
  }
  return 0;
}
int calculate_fileread(const char* filename){
  char * line = NULL;
  size_t len = 0;
  ssize_t read;
  FILE *fptr;  
  numberofsources=0;
  fptr = fopen(filename, "r");
  if(fptr == NULL)
  {
     g_print("File read error!\n"); 
     return -1;
  }
  
  //set reference boxes
  
  int linenum=-1;
  while ((read = getline(&line, &len, fptr)) != -1) {
     char * token = strtok(line, ":");
     int count=0;
     char * current;
     while( token != NULL ) { 
        if(count==0 ){
          current=token;          
        }
        if(strcmp(current, "numsources") == 0 && count>0)
        {
            numberofsources = atoi(token);            
            refboxes_calculate = (refbox*)malloc(numberofsources * sizeof(refbox)); 
           
        }   
        else if(strstr(current, "source") != NULL && count>0)
        {
           //printf("\nlinenum %d\n",linenum);
           //printf("current %s\n",current);
           char * source1 = strtok(current, "source"); 
           //printf("source1 %s\n",source1);
           char * source2 = strtok(source1, "-");           
           refboxes_calculate[linenum].source_num = atoi(source2)-1;
           //printf("source_num %d\n", refboxes_calculate[linenum].source_num);
           while( source2 != NULL ) { 
              source2 = strtok(NULL, " ");
              refboxes_calculate[linenum].size = atoi(source2)*4;
              refboxes_calculate[linenum].corners = (int*)malloc(atoi(source2)*4 * sizeof(int));
              source2 = strtok(NULL, " ");
           }
           int countcorners=0;
           //printf("\ntoken %s\n",token);
	   do{  
		int l = strcspn (token, ",");
                //printf("countcorners %d num %d ",countcorners, atoi(token));
		refboxes_calculate[linenum].corners[countcorners] = atoi(token);  
		token += l + 1; 
		countcorners = countcorners+1; 		      
	  }while(token[-1]);
        }        
          
        token = strtok(NULL, " ");
        count=count+1;
     }
     linenum=linenum+1;
  }
  
  fclose(fptr);
  printf("numberofsources %d\n ", numberofsources);
  for(linenum=0; linenum < numberofsources; linenum++){
     printf( "\nsources %d size %d\n", refboxes_calculate[linenum].source_num, refboxes_calculate[linenum].size );
     for(int countcorners=0; countcorners < refboxes_calculate[linenum].size; countcorners++)
       printf( "%d ", refboxes_calculate[linenum].corners[countcorners] );
  }
  return 0;
}
#endif
/**
 * Initialize all resources and start the output thread
 */
static gboolean
gst_dsexample_start (GstBaseTransform * btrans)
{
  GstDsExample *dsexample = GST_DSEXAMPLE (btrans);
  NvBufSurfaceCreateParams create_params;
  DsExampleInitParams init_params =
      { dsexample->processing_width, dsexample->processing_height,
    dsexample->process_full_frame
  };
#ifdef ENABLE_MAPLETREE
  display_fileread("/opt/nvidia/deepstream/deepstream-5.1/sources/gst-plugins/gst-dsexample/display_rois.txt");
  calculate_fileread("/opt/nvidia/deepstream/deepstream-5.1/sources/gst-plugins/gst-dsexample/calculate_rois.txt");
  // Creating a new thread
  threadrunning=TRUE;
  pthread_create(&ptid, NULL, &func, NULL);
  printf("\n******************Thread started*******************");
#endif
  GstQuery *queryparams = NULL;
  guint batch_size = 1;

  /* Algorithm specific initializations and resource allocation. */
  dsexample->dsexamplelib_ctx = DsExampleCtxInit (&init_params);

  GST_DEBUG_OBJECT (dsexample, "ctx lib %p \n", dsexample->dsexamplelib_ctx);

  CHECK_CUDA_STATUS (cudaSetDevice (dsexample->gpu_id),
      "Unable to set cuda device");

  dsexample->batch_size = 1;
  queryparams = gst_nvquery_batch_size_new ();
  if (gst_pad_peer_query (GST_BASE_TRANSFORM_SINK_PAD (btrans), queryparams)
      || gst_pad_peer_query (GST_BASE_TRANSFORM_SRC_PAD (btrans), queryparams)) {
    if (gst_nvquery_batch_size_parse (queryparams, &batch_size)) {
      dsexample->batch_size = batch_size;
    }
  }
  GST_DEBUG_OBJECT (dsexample, "Setting batch-size %d \n",
      dsexample->batch_size);
  gst_query_unref (queryparams);

  if (dsexample->process_full_frame && dsexample->blur_objects) {
    GST_ERROR ("Error: does not support blurring while processing full frame");
    goto error;
  }

  CHECK_CUDA_STATUS (cudaStreamCreate (&dsexample->cuda_stream),
      "Could not create cuda stream");

  if (dsexample->inter_buf)
    NvBufSurfaceDestroy (dsexample->inter_buf);
  dsexample->inter_buf = NULL;

  /* An intermediate buffer for NV12/RGBA to BGR conversion  will be
   * required. Can be skipped if custom algorithm can work directly on NV12/RGBA. */
  create_params.gpuId  = dsexample->gpu_id;
  create_params.width  = dsexample->processing_width;
  create_params.height = dsexample->processing_height;
  create_params.size = 0;
  create_params.colorFormat = NVBUF_COLOR_FORMAT_RGBA;
  create_params.layout = NVBUF_LAYOUT_PITCH;
#ifdef __aarch64__
  create_params.memType = NVBUF_MEM_DEFAULT;
#else
  create_params.memType = NVBUF_MEM_CUDA_UNIFIED;
#endif

  if (NvBufSurfaceCreate (&dsexample->inter_buf, 1,
          &create_params) != 0) {
    GST_ERROR ("Error: Could not allocate internal buffer for dsexample");
    goto error;
  }

  /* Create host memory for storing converted/scaled interleaved RGB data */
  CHECK_CUDA_STATUS (cudaMallocHost (&dsexample->host_rgb_buf,
          dsexample->processing_width * dsexample->processing_height *
          RGB_BYTES_PER_PIXEL), "Could not allocate cuda host buffer");

  GST_DEBUG_OBJECT (dsexample, "allocated cuda buffer %p \n",
      dsexample->host_rgb_buf);

  /* CV Mat containing interleaved RGB data. This call does not allocate memory.
   * It uses host_rgb_buf as data. */
  dsexample->cvmat =
      new cv::Mat (dsexample->processing_height, dsexample->processing_width,
      CV_8UC3, dsexample->host_rgb_buf,
      dsexample->processing_width * RGB_BYTES_PER_PIXEL);

  if (!dsexample->cvmat)
    goto error;

  GST_DEBUG_OBJECT (dsexample, "created CV Mat\n");

  return TRUE;
error:
  if (dsexample->host_rgb_buf) {
    cudaFreeHost (dsexample->host_rgb_buf);
    dsexample->host_rgb_buf = NULL;
  }

  if (dsexample->cuda_stream) {
    cudaStreamDestroy (dsexample->cuda_stream);
    dsexample->cuda_stream = NULL;
  }
  if (dsexample->dsexamplelib_ctx)
    DsExampleCtxDeinit (dsexample->dsexamplelib_ctx);

  return FALSE;
}

/**
 * Stop the output thread and free up all the resources
 */
static gboolean
gst_dsexample_stop (GstBaseTransform * btrans)
{
  GstDsExample *dsexample = GST_DSEXAMPLE (btrans);

  if (dsexample->inter_buf)
    NvBufSurfaceDestroy(dsexample->inter_buf);
  dsexample->inter_buf = NULL;

  if (dsexample->cuda_stream)
    cudaStreamDestroy (dsexample->cuda_stream);
  dsexample->cuda_stream = NULL;

  delete dsexample->cvmat;
  dsexample->cvmat = NULL;

  if (dsexample->host_rgb_buf) {
    cudaFreeHost (dsexample->host_rgb_buf);
    dsexample->host_rgb_buf = NULL;
  }
#ifdef ENABLE_MAPLETREE
  for(int source=0; source < numberofsources; source++){
       free( refboxes_display[source].corners );
       free( refboxes_calculate[source].corners );
  }
  free(refboxes_display); free(refboxes_calculate);
  std::vector<box>().swap(boxes);
  std::vector<box>().swap(illegalparked_boxes);
  // Waiting for the created thread to terminate
  threadrunning=FALSE;
  pthread_join(ptid, NULL);
  printf("******************Thread ended*******************\n");
#endif
  GST_DEBUG_OBJECT (dsexample, "deleted CV Mat \n");

  /* Deinit the algorithm library */
  DsExampleCtxDeinit (dsexample->dsexamplelib_ctx);
  dsexample->dsexamplelib_ctx = NULL;

  GST_DEBUG_OBJECT (dsexample, "ctx lib released \n");

  return TRUE;
}

/**
 * Called when source / sink pad capabilities have been negotiated.
 */
static gboolean
gst_dsexample_set_caps (GstBaseTransform * btrans, GstCaps * incaps,
    GstCaps * outcaps)
{
  GstDsExample *dsexample = GST_DSEXAMPLE (btrans);
  /* Save the input video information, since this will be required later. */
  gst_video_info_from_caps (&dsexample->video_info, incaps);

  if (dsexample->blur_objects && !dsexample->process_full_frame) {
    /* requires RGBA format for blurring the objects in opencv */
     if (dsexample->video_info.finfo->format != GST_VIDEO_FORMAT_RGBA) {
      GST_ELEMENT_ERROR (dsexample, STREAM, FAILED,
          ("input format should be RGBA when using blur-objects property"), (NULL));
      goto error;
      }
  }

  return TRUE;

error:
  return FALSE;
}

/**
 * Scale the entire frame to the processing resolution maintaining aspect ratio.
 * Or crop and scale objects to the processing resolution maintaining the aspect
 * ratio. Remove the padding required by hardware and convert from RGBA to RGB
 * using openCV. These steps can be skipped if the algorithm can work with
 * padded data and/or can work with RGBA.
 */
static GstFlowReturn
get_converted_mat (GstDsExample * dsexample, NvBufSurface *input_buf, gint idx, NvOSD_RectParams * crop_rect_params, gdouble & ratio, gint input_width, gint input_height)
{
if(FALSE)
{
  NvBufSurfTransform_Error err;
  NvBufSurfTransformConfigParams transform_config_params;
  NvBufSurfTransformParams transform_params;
  NvBufSurfTransformRect src_rect;
  NvBufSurfTransformRect dst_rect;
  NvBufSurface ip_surf;
  cv::Mat in_mat;
  ip_surf = *input_buf;
  ip_surf.numFilled = ip_surf.batchSize = 1;
  ip_surf.surfaceList = &(input_buf->surfaceList[idx]);

  gint src_left = GST_ROUND_UP_2((unsigned int)crop_rect_params->left);
  gint src_top = GST_ROUND_UP_2((unsigned int)crop_rect_params->top);
  gint src_width = GST_ROUND_DOWN_2((unsigned int)crop_rect_params->width);
  gint src_height = GST_ROUND_DOWN_2((unsigned int)crop_rect_params->height);

  /* Maintain aspect ratio */
  double hdest = dsexample->processing_width * src_height / (double) src_width;
  double wdest = dsexample->processing_height * src_width / (double) src_height;
  guint dest_width, dest_height;

  if (hdest <= dsexample->processing_height) {
    dest_width = dsexample->processing_width;
    dest_height = hdest;
  } else {
    dest_width = wdest;
    dest_height = dsexample->processing_height;
  }

  /* Configure transform session parameters for the transformation */
  transform_config_params.compute_mode = NvBufSurfTransformCompute_Default;
  transform_config_params.gpu_id = dsexample->gpu_id;
  transform_config_params.cuda_stream = dsexample->cuda_stream;

  /* Set the transform session parameters for the conversions executed in this
   * thread. */
  err = NvBufSurfTransformSetSessionParams (&transform_config_params);
  if (err != NvBufSurfTransformError_Success) {
    GST_ELEMENT_ERROR (dsexample, STREAM, FAILED, ("NvBufSurfTransformSetSessionParams failed with error %d", err), (NULL));
    goto error;
  }

  /* Calculate scaling ratio while maintaining aspect ratio */
  ratio = MIN (1.0 * dest_width/ src_width, 1.0 * dest_height / src_height);

  if ((crop_rect_params->width == 0) || (crop_rect_params->height == 0)) {
    GST_ELEMENT_ERROR (dsexample, STREAM, FAILED, ("%s:crop_rect_params dimensions are zero",__func__), (NULL));
    goto error;
  }

#ifdef __aarch64__
  if (ratio <= 1.0 / 16 || ratio >= 16.0) {
    /* Currently cannot scale by ratio > 16 or < 1/16 for Jetson */
    goto error;
  }
#endif
  /* Set the transform ROIs for source and destination */
  src_rect = {(guint)src_top, (guint)src_left, (guint)src_width, (guint)src_height};
  dst_rect = {0, 0, (guint)dest_width, (guint)dest_height};

  /* Set the transform parameters */
  transform_params.src_rect = &src_rect;
  transform_params.dst_rect = &dst_rect;
  transform_params.transform_flag = NVBUFSURF_TRANSFORM_FILTER | NVBUFSURF_TRANSFORM_CROP_SRC | NVBUFSURF_TRANSFORM_CROP_DST;
  transform_params.transform_filter = NvBufSurfTransformInter_Default;

  /* Memset the memory */
  NvBufSurfaceMemSet (dsexample->inter_buf, 0, 0, 0);

  GST_DEBUG_OBJECT (dsexample, "Scaling and converting input buffer\n");

  /* Transformation scaling+format conversion if any. */
  err = NvBufSurfTransform (&ip_surf, dsexample->inter_buf, &transform_params);
  if (err != NvBufSurfTransformError_Success) {
    GST_ELEMENT_ERROR (dsexample, STREAM, FAILED, ("NvBufSurfTransform failed with error %d while converting buffer", err),
        (NULL));
    goto error;
  }
  /* Map the buffer so that it can be accessed by CPU */
  if (NvBufSurfaceMap (dsexample->inter_buf, 0, 0, NVBUF_MAP_READ) != 0){
    goto error;
  }

  /* Cache the mapped data for CPU access */
  NvBufSurfaceSyncForCpu (dsexample->inter_buf, 0, 0);

  /* Use openCV to remove padding and convert RGBA to BGR. Can be skipped if
   * algorithm can handle padded RGBA data. */
  in_mat = cv::Mat (dsexample->processing_height, dsexample->processing_width, CV_8UC4, dsexample->inter_buf->surfaceList[0].mappedAddr.addr[0],  dsexample->inter_buf->surfaceList[0].pitch);

#if (CV_MAJOR_VERSION >= 4)
  cv::cvtColor (in_mat, *dsexample->cvmat, cv::COLOR_RGBA2BGR);
#else
  cv::cvtColor (in_mat, *dsexample->cvmat, CV_RGBA2BGR);
#endif

  if (NvBufSurfaceUnMap (dsexample->inter_buf, 0, 0)){
    goto error;
  }

#ifdef __aarch64__
  /* To use the converted buffer in CUDA, create an EGLImage and then use
   * CUDA-EGL interop APIs */
  if (USE_EGLIMAGE) {
    if (NvBufSurfaceMapEglImage (dsexample->inter_buf, 0) !=0 ) {
      goto error;
    }

    /* dsexample->inter_buf->surfaceList[0].mappedAddr.eglImage
     * Use interop APIs cuGraphicsEGLRegisterImage and
     * cuGraphicsResourceGetMappedEglFrame to access the buffer in CUDA */

    /* Destroy the EGLImage */
    NvBufSurfaceUnMapEglImage (dsexample->inter_buf, 0);
  }
#endif

  /* We will first convert only the Region of Interest (the entire frame or the
   * object bounding box) to RGB and then scale the converted RGB frame to
   * processing resolution. */
}
  return GST_FLOW_OK;

error:
  return GST_FLOW_ERROR;
}

/**
 * Called when element recieves an input buffer from upstream element.
 */
static GstFlowReturn
gst_dsexample_transform_ip (GstBaseTransform * btrans, GstBuffer * inbuf)
{  
  GstFlowReturn flow_ret = GST_FLOW_ERROR;
  GstDsExample *dsexample = GST_DSEXAMPLE (btrans);
  GstMapInfo in_map_info;
  
  gdouble scale_ratio = 1.0;
  DsExampleOutput *output;

  NvBufSurface *surface = NULL;
  NvDsBatchMeta *batch_meta = NULL;
  NvDsFrameMeta *frame_meta = NULL;
  NvDsMetaList * l_frame = NULL;
  guint i = 0;
  /*Plot mask*/
  NvBufSurface* roi_surface;
  NvBufSurface* color_surface;
  NvBufSurface* alpha_surface;
  NvBufSurfaceCreateParams create_params;
  create_params.gpuId  = dsexample->gpu_id;  
  create_params.size = 0;
  create_params.colorFormat = NVBUF_COLOR_FORMAT_RGBA;
  create_params.layout = NVBUF_LAYOUT_PITCH;
  _NvBufSurfTransformCompositeBlendParams *params;
  //create_params.memType = NVBUF_MEM_DEFAULT;
  create_params.memType = NVBUF_MEM_CUDA_UNIFIED;


  /*Plot mask*/
  dsexample->frame_num++;
  CHECK_CUDA_STATUS (cudaSetDevice (dsexample->gpu_id),
      "Unable to set cuda device");

  memset (&in_map_info, 0, sizeof (in_map_info));
  if (!gst_buffer_map (inbuf, &in_map_info, GST_MAP_READ)) {
    g_print ("Error: Failed to map gst buffer\n");
    goto error;
  }

  nvds_set_input_system_timestamp (inbuf, GST_ELEMENT_NAME (dsexample));
  surface = (NvBufSurface *) in_map_info.data;
  GST_DEBUG_OBJECT (dsexample,
      "Processing Frame %" G_GUINT64_FORMAT " Surface %p\n",
      dsexample->frame_num, surface);

  if (CHECK_NVDS_MEMORY_AND_GPUID (dsexample, surface))
    goto error;

  batch_meta = gst_buffer_get_nvds_batch_meta (inbuf);
  if (batch_meta == nullptr) {
    GST_ELEMENT_ERROR (dsexample, STREAM, FAILED,
        ("NvDsBatchMeta not found for input buffer."), (NULL));
    return GST_FLOW_ERROR;
  }
#ifdef ENABLE_MAPLETREE
  if (dsexample->process_full_frame) {
    for (l_frame = batch_meta->frame_meta_list; l_frame != NULL; l_frame = l_frame->next)
    {      
        frame_meta = (NvDsFrameMeta *) (l_frame->data);        
        if (surface->surfaceList[frame_meta->batch_id].mappedAddr.addr[0] == NULL){
           if (NvBufSurfaceMap (surface, frame_meta->batch_id, 0, NVBUF_MAP_READ_WRITE) != 0){
                GST_ELEMENT_ERROR (dsexample, STREAM, FAILED,
                ("%s:buffer map to be accessed by CPU failed", __func__), (NULL));
                return GST_FLOW_ERROR;
           }
        }
        for(int source=0; source < numberofsources; source++){
           guint stream_id = frame_meta->pad_index;
           int width = frame_meta->source_frame_width;
           int height = frame_meta->source_frame_height;  
           //printf("frame_meta->source_id %d refboxes_display[source].source_num %d \n", frame_meta->source_id, refboxes_display[source].source_num);
           if(frame_meta->source_id == refboxes_display[source].source_num){                
                //printf("plot at %d w %d h %d\n",refboxes_display[source].source_num, surface->surfaceList[frame_meta->batch_id].width, surface->surfaceList[frame_meta->batch_id].height );
	        NvBufSurfaceSyncForCpu (surface, frame_meta->batch_id, 0);
                cv::Mat in_mat = cv::Mat (surface->surfaceList[frame_meta->batch_id].height, surface->surfaceList[frame_meta->batch_id].width, CV_8UC4,
		                  surface->surfaceList[frame_meta->batch_id].mappedAddr.addr[0], surface->surfaceList[frame_meta->batch_id].pitch);
                for(int boxidx=0; boxidx < refboxes_display[source].size; boxidx=boxidx+4){
                       //printf("w %d h %d xmin %d ymin %d xmax %d ymax %d\n",surface->surfaceList[frame_meta->batch_id].planeParams.width[0], 
                       //surface->surfaceList[frame_meta->batch_id].planeParams.height[0], refboxes_display[source].corners[boxidx+0], 
                       //refboxes_display[source].corners[boxidx+1], refboxes_display[source].corners[boxidx+2], refboxes_display[source].corners[boxidx+3]);                       
                       int x1 = refboxes_display[source].corners[boxidx+0];  int y1 = refboxes_display[source].corners[boxidx+1];
                       int x2 = refboxes_display[source].corners[boxidx+2];  int y2 = refboxes_display[source].corners[boxidx+3];
                       create_params.width  = x2 - x1;
                       create_params.height = y2 - y1;
                       
                       std::lock_guard<std::mutex> guard(alert_lock);
                       if(alertsoruce[source] == 1){
                          cv::Mat roi_alert;
                          if(source == 0)
                             roi_alert = in_mat(cv::Rect(0, 0, 480, 1080)); 
                          else if(source == 1)
                             roi_alert = in_mat(cv::Rect(0, 0, 480, 1080));
                          else if(source == 2)
                             roi_alert = in_mat(cv::Rect(0, 0, 480, 1080));
                          else if(source == 3)
                             roi_alert = in_mat(cv::Rect(1025, 0, 1505, 1080));
                          cv::Mat color_alert(roi_alert.size(), CV_8UC4, cv::Scalar( 0, 0, 0 ));
                          cv::addWeighted(color_alert, 0.5, roi_alert, 1.0 - 0.5, 0.0, roi_alert);
                       }
                       //cv::rectangle( in_mat, cv::Point( x1, y1 ), cv::Point( x2, y2 ), cv::Scalar( 0, 128, 0, 0.1 ), cv::FILLED);
                       cv::Mat roi = in_mat(cv::Rect(x1, y1, x2-x1, y2-y1 )); 
                       cv::Mat color(roi.size(), CV_8UC4, cv::Scalar( 0, 0, 255, 0 ));
                       cv::Mat alpha(roi.size(), CV_8UC4, cv::Scalar( 1, 1, 1, 1 ));
                       //double alpha = 0.3;
                       //cv::addWeighted(color, alpha, roi, 1.0 - alpha, 0.0, roi);
                       if (NvBufSurfaceCreate (&roi_surface, 1, &create_params) != 0) {
			    printf ("Error: in Create\n");
		       }
		       //NvBufSurfaceMemSet (roi_surface, 0, 0, 0);

		       if (NvBufSurfaceMap (roi_surface, 0, 0, NVBUF_MAP_READ_WRITE) != 0){
			    printf("Err in Map\n");
		       }
		       if (NvBufSurfaceSyncForCpu (roi_surface, 0, 0) !=0) {
			    printf("Err in Synccpu");
		       }
		       memcpy(roi_surface->surfaceList[0].mappedAddr.addr[0], roi.ptr(), create_params.width * create_params.height * 4);
                       if (NvBufSurfaceCreate (&color_surface, 1, &create_params) != 0) {
			    printf ("Error: in Create\n");
		       }
		       //NvBufSurfaceMemSet (color_surface, 0, 0, 0);

		       if (NvBufSurfaceMap (color_surface, 0, 0, NVBUF_MAP_READ_WRITE) != 0){
			    printf("Err in Map\n");
		       }
		       if (NvBufSurfaceSyncForCpu (color_surface, 0, 0) !=0) {
			    printf("Err in Synccpu");
		       }
		       memcpy(color_surface->surfaceList[0].mappedAddr.addr[0], color.ptr(), create_params.width * create_params.height * 4);
                       if (NvBufSurfaceCreate (&alpha_surface, 1, &create_params) != 0) {
			    printf ("Error: in Create\n");
		       }
		       //NvBufSurfaceMemSet (alpha_surface, 0, 0, 0);

		       if (NvBufSurfaceMap (alpha_surface, 0, 0, NVBUF_MAP_READ_WRITE) != 0){
			    printf("Err in Map\n");
		       }
		       if (NvBufSurfaceSyncForCpu (alpha_surface, 0, 0) !=0) {
			    printf("Err in Synccpu");
		       }
		       memcpy(alpha_surface->surfaceList[0].mappedAddr.addr[0], alpha.ptr(), create_params.width * create_params.height * 4);
                       
                       //cv::putText(in_mat, std::to_string(x1)+','+std::to_string(y1), cv::Point(x1,y1), cv::FONT_HERSHEY_SIMPLEX, 1.0, cv::Scalar(0,0,0), 1);
                       //cv::Mat roi = in_mat(cv::Rect(x1, y1, x2-x1, y2-y1 )); 
                       //cv::Mat color(roi.size(), CV_8UC4, cv::Scalar( 255, 0, 255 ));
                       //double alpha = 0.3;
                       
                       NvBufSurfTransformCompositeBlend(roi_surface, color_surface, alpha_surface, roi_surface, params);
                       
                     
                }
                for (NvDsMetaList * l_obj = frame_meta->obj_meta_list; l_obj != NULL; l_obj = l_obj->next) {
		     NvDsObjectMeta *obj = (NvDsObjectMeta *) l_obj->data;
		     int left = (int)obj->tracker_bbox_info.org_bbox_coords.left;
		     int top = (int)obj->tracker_bbox_info.org_bbox_coords.top;
		     int right = (int)left + obj->tracker_bbox_info.org_bbox_coords.width;
		     int bottom = (int)top + obj->tracker_bbox_info.org_bbox_coords.height;

		     box box_;
		     box_.source_num = stream_id; box_.id = obj->object_id; gettimeofday(&box_.start_tv, NULL);
		     box_.x_min = (int)left; box_.y_min = (int)top; box_.x_max = (int)right; box_.y_max = (int)bottom;
		     std::lock_guard<std::mutex> guard(m_lock);
                     boxes.push_back(box_);
 		     //g_print("\n%s %d %lu %d %d %d %d ","input ", stream_id, box_.id, box_.x_min, box_.y_min, box_.x_max, box_.y_max);

		}
                
                break;
	   }  
        }
    } 


  } 
#endif
  flow_ret = GST_FLOW_OK;

error:

  nvds_set_output_system_timestamp (inbuf, GST_ELEMENT_NAME (dsexample));
  gst_buffer_unmap (inbuf, &in_map_info);

  return flow_ret;
}

/**
 * Attach metadata for the full frame. We will be adding a new metadata.
 */
static void
attach_metadata_full_frame (GstDsExample * dsexample, NvDsFrameMeta *frame_meta,
    gdouble scale_ratio, DsExampleOutput * output, guint batch_id)
{
  NvDsBatchMeta *batch_meta = frame_meta->base_meta.batch_meta;
  NvDsObjectMeta *object_meta = NULL;
  static gchar font_name[] = "Serif";
  GST_DEBUG_OBJECT (dsexample, "Attaching metadata %d\n", output->numObjects);

  for (gint i = 0; i < output->numObjects; i++) {
    DsExampleObject *obj = &output->object[i];
    object_meta = nvds_acquire_obj_meta_from_pool(batch_meta);
    NvOSD_RectParams & rect_params = object_meta->rect_params;
    NvOSD_TextParams & text_params = object_meta->text_params;

    /* Assign bounding box coordinates */
    rect_params.left = obj->left;
    rect_params.top = obj->top;
    rect_params.width = obj->width;
    rect_params.height = obj->height;

    /* Semi-transparent yellow background */
    rect_params.has_bg_color = 0;
    rect_params.bg_color = (NvOSD_ColorParams) {
    1, 1, 0, 0.4};
    /* Red border of width 6 */
    rect_params.border_width = 3;
    rect_params.border_color = (NvOSD_ColorParams) {
    1, 0, 0, 1};

    /* Scale the bounding boxes proportionally based on how the object/frame was
     * scaled during input */
    rect_params.left /= scale_ratio;
    rect_params.top /= scale_ratio;
    rect_params.width /= scale_ratio;
    rect_params.height /= scale_ratio;
    GST_DEBUG_OBJECT (dsexample, "Attaching rect%d of batch%u"
        "  left->%f top->%f width->%f"
        " height->%f label->%s\n", i, batch_id, rect_params.left,
        rect_params.top, rect_params.width, rect_params.height, obj->label);

    object_meta->object_id = UNTRACKED_OBJECT_ID;
    g_strlcpy (object_meta->obj_label, obj->label, MAX_LABEL_SIZE);
    /* display_text required heap allocated memory */
    text_params.display_text = g_strdup (obj->label);
    /* Display text above the left top corner of the object */
    text_params.x_offset = rect_params.left;
    text_params.y_offset = rect_params.top - 10;
    /* Set black background for the text */
    text_params.set_bg_clr = 1;
    text_params.text_bg_clr = (NvOSD_ColorParams) {
    0, 0, 0, 1};
    /* Font face, size and color */
    text_params.font_params.font_name = font_name;
    text_params.font_params.font_size = 11;
    text_params.font_params.font_color = (NvOSD_ColorParams) {
    1, 1, 1, 1};

    nvds_add_obj_meta_to_frame(frame_meta, object_meta, NULL);
    frame_meta->bInferDone = TRUE;
  }
}

/**
 * Only update string label in an existing object metadata. No bounding boxes.
 * We assume only one label per object is generated
 */
static void
attach_metadata_object (GstDsExample * dsexample, NvDsObjectMeta * obj_meta,
    DsExampleOutput * output)
{
  if (output->numObjects == 0)
    return;
  NvDsBatchMeta *batch_meta = obj_meta->base_meta.batch_meta;

  NvDsClassifierMeta *classifier_meta =
    nvds_acquire_classifier_meta_from_pool (batch_meta);

  classifier_meta->unique_component_id = dsexample->unique_id;

  NvDsLabelInfo *label_info =
    nvds_acquire_label_info_meta_from_pool (batch_meta);
  g_strlcpy (label_info->result_label, output->object[0].label, MAX_LABEL_SIZE);
  nvds_add_label_info_meta_to_classifier(classifier_meta, label_info);
  nvds_add_classifier_meta_to_object (obj_meta, classifier_meta);

  nvds_acquire_meta_lock (batch_meta);
  NvOSD_TextParams & text_params = obj_meta->text_params;
  NvOSD_RectParams & rect_params = obj_meta->rect_params;

  /* Below code to display the result */
  /* Set black background for the text
   * display_text required heap allocated memory */
  if (text_params.display_text) {
    gchar *conc_string = g_strconcat (text_params.display_text, " ",
        output->object[0].label, NULL);
    g_free (text_params.display_text);
    text_params.display_text = conc_string;
  } else {
    /* Display text above the left top corner of the object */
    text_params.x_offset = rect_params.left;
    text_params.y_offset = rect_params.top - 10;
    text_params.display_text = g_strdup (output->object[0].label);
    /* Font face, size and color */
    text_params.font_params.font_name = (char *)"Serif";
    text_params.font_params.font_size = 11;
    text_params.font_params.font_color = (NvOSD_ColorParams) {
    1, 1, 1, 1};
    /* Set black background for the text */
    text_params.set_bg_clr = 1;
    text_params.text_bg_clr = (NvOSD_ColorParams) {
    0, 0, 0, 1};
  }
  nvds_release_meta_lock (batch_meta);
}

/**
 * Boiler plate for registering a plugin and an element.
 */
static gboolean
dsexample_plugin_init (GstPlugin * plugin)
{
  GST_DEBUG_CATEGORY_INIT (gst_dsexample_debug, "dsexample", 0,
      "dsexample plugin");

  return gst_element_register (plugin, "dsexample", GST_RANK_PRIMARY,
      GST_TYPE_DSEXAMPLE);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    nvdsgst_dsexample,
    DESCRIPTION, dsexample_plugin_init, DS_VERSION, LICENSE, BINARY_PACKAGE, URL)
